"""
A simplified interface to Windows MIDI services.

You will need to understand the MIDI byte protocol, a summary of which
can be found at <http://www.midi.org/about-midi/table1.shtml>

There are limitations of the underlying Win32 MIDI API. Some of these
can be found at <http://msdn.microsoft.com/library/en-us/multimed/htm/_win32_musical_instrument_digital_interface_midi_.asp>
"""

__author__ = "David Leonard"
__copyright__ = "Copyright 2004, Adaptive Enterprises"
__version__ = 0.2

import exceptions
import time
import _winmm_midi
_w = _winmm_midi

#-- timeout constant used in InputDevice.read()
NoTimeout = _w.INFINITE

#-- msg types from InputDevice.read()
Data =      _w.MIM_DATA
LongData =  _w.MIM_LONGDATA
MoreData =  _w.MIM_MOREDATA
Error =     _w.MIM_ERROR
LongError = _w.MIM_LONGERROR

#-- values for {Input,Output}Device.state
Closed  =   intern('Closed')
Stopped =   intern('Stopped')
Started =   intern('Started')

#-- values for {Input,Output}Device.supported
Cache =	    intern('Cache')
Volume =    intern('Volume')
LRVolume =  intern('LRVolume')
Stream =    intern('Stream')

#-- values for OutputDevice.technology
MidiPort =  _w.MOD_MIDIPORT
Synth =     _w.MOD_SYNTH
SQSynth =   _w.MOD_SQSYNTH
FMSynth =   _w.MOD_FMSYNTH
Mapper =    _w.MOD_MAPPER
WaveTable = 6 # _w.MOD_WAVETABLE
SWSynth =   7 # _w.MOD_SWSYNTH

#-- values for {Input,Output}Device.manufacturer_id
Microsoft = 1

#
# Exception classes
#
class Exception(exceptions.Exception): pass
class Timeout(Exception): pass
class NoMatch(Exception): pass

class Error(Exception):
	def __init__(self, mmresult, func, text):
	    self.mmresult = mmresult
	    exceptions.Exception.__init__(self,
		"%s (error #%d from '%s()')" % (text, mmresult, func))
class BadDeviceIdError(Error): pass
class NotEnabledError(Error): pass
class AllocatedError(Error): pass
class InvalHandleError(Error): pass
class NoDriverError(Error): pass
class NoMemError(Error): pass
class NotSupportedError(Error): pass
class BadErrNumError(Error): pass
class InvalFlagError(Error): pass
class InvalParamError(Error): pass
class HandleBusyError(Error): pass
class InvalidAliasError(Error): pass
class BadDBError(Error): pass
class KeyNotFoundError(Error): pass
class ReadError(Error): pass
class WriteError(Error): pass
class DeleteError(Error): pass
class ValNotFound(Error): pass
class NoDriverCBError(Error): pass
class UnknownError(Error): pass

#-- create a mapping from MMSYSERR_* numbers to the Exception classes
_errormap = {}
for errorclass in (Error, BadDeviceIdError, NotEnabledError, AllocatedError,
		   InvalHandleError, NoDriverError, NoMemError,
		   NotSupportedError, BadErrNumError, InvalFlagError,
		   InvalParamError, HandleBusyError, InvalidAliasError,
		   BadDBError, KeyNotFoundError, ReadError, WriteError,
		   DeleteError, ValNotFound, NoDriverCBError):
    constname = "MMSYSERR_" + errorclass.__name__.upper()
    if not hasattr(_w, constname): constname = constname[:-5] # trim 'Error'
    _errormap[getattr(_w, constname)] = errorclass

#-- convenience exception constructor
def _mmerror(mmresult, func, isInput):
    """_mmerror(mmresult, isInput) -> Error
       Creates a new error object from an MMRESULT error number."""
    if isInput:
        mmr2,text = _w.midiInGetErrorText(mmresult, _w.MAXERRORLENGTH)
    else:
	mmr2,text = _w.midiOutGetErrorText(mmresult, _w.MAXERRORLENGTH)
    if mmr2 or not _errormap.has_key(mmresult):
	return UnknownError(mmresult, func, "unknown error")
    else:
	return _errormap[mmresult](mmresult, func, text)


class InputDevice:
    """A MIDI input device.
       An input device has the following properties:
	   .id		    - the system index
           .name	    - the product name 
	   .manufacturer_id - the manufacturer's ID
	   .product_id      - the manufacturer's product ID
	   .driver_version  - the (major,minor) tuple of the installed driver
	   .supported 	    - a tuple of zero or more supported capabilities:
	   			Cache, LRVolume, Stream, Volume
	   .isopen	    - readonly boolean indicating open() was called
	   .state	    - readonly value indicate device soft state
	   			Closed, Stopped, Started
           .nocheck	    - Treat input data marked 'error' as normal data
       Calling open() on a device is optional; most methods will call it
       automatically if the isopen property is false.
    """


    def __init__(self, id, caps):
	self.id = id

	#-- use capabilities dictionary to initialise some properties
	self.name =            caps['pname']
	self.manufacturer_id = caps['mid']
	self.product_id =      caps['pid']
	self.driver_version =  caps['driverVersion']
	_sup = []
        for nm in (Cache, LRVolume, Stream, Volume):
	    if caps["support"] & getattr(_w, 'MIDICAPS_'+nm.upper()):
		_sup.append(nm)
	self.supported = tuple(_sup)

	#-- initialise self to use no resources
	self.isopen = False
	self.state = Closed
	self._q = None
	self._midiIn = None
	self._hdr = None
	self._readq = []
	self.nocheck = False

    def __repr__(self):
	return "<InputDevice id %d state %s>" % (
		self.id, self.state)

    def __cmp__(self, other):
	"""Compare two midi input devices"""
	if isinstance(other, InputDevice):
	    return cmp((self.id, self.state), (other.id, other.state))
	else:
	    return cmp(id(self), id(other))

    def open(self, buffersz=8192):
        """open() -> None
           Opens the device for input, and places it in the Stopped state."""
	assert not self.isopen

	self._q = _w.new_MidiMsgQueue(buffersz)
	mmresult = _w.midiInOpen(self.id, self._q)
	if type(mmresult) is type(0):
	    raise _mmerror(mmresult, "midiInOpen", True)
	self._midiIn = mmresult[0]

	self.isopen = True
	self.state = Stopped

	try:
	    self._hdr = _w.new_MIDIHDR(buffersz)
	    mmresult = _w.midiInPrepareHeader(self._midiIn, self._hdr)
	    if mmresult: raise _mmerror(mmresult, "midiInPrepareHeader", True)
	    mmresult = _w.midiInAddBuffer(self._midiIn, self._hdr)
	    if mmresult: raise _mmerror(mmresult, "midiInAddBuffer", True)
	except:
	    # close on any error
	    self.close()
	    raise

    def start(self):
	"""start() -> None
	   Starts input on this MIDI device. Internally, this function resets 
	   the time stamp to zero; time stamp values for subsequently received 
	   messages are relative to the time that this function was called.  
	   """
	if not self.isopen: self.open()
	assert self.isopen

	mmresult = _w.midiInStart(self._midiIn)
	if mmresult: raise _mmerror(mmresult, "midiInStart", True)
	self.state = Started

    def stop(self):
	"""stop() -> None
	   Stops input on this midi device. Internally, this function returns 
	   all pending input buffers to the callback function and sets the 
	   MHDR_DONE flag in the dwFlags member of the MIDIHDR structure."""
	assert self.isopen

	mmresult = _w.midiInStop(self._midiIn)
	if mmresult: raise _mmerror(mmresult, "midiInStop", True)
	self.state = Stopped

    def reset(self):
	"""reset() -> None
	   Resets the midi device"""
	assert self.isopen

	#-- turn off header recycling
	_w.MidiMsgQueue_recycle_set(self._q, False)
	mmresult = _w.midiInReset(self._midiIn)
	if mmresult: raise _mmerror(mmresult, "midiInReset", True)

	#-- wait a bit for any driver sluggishness
	time.sleep(0.5)
	self.state = Stopped
	_w.MidiMsgQueue_read(self._q, 0) # result ignored
	_w.MidiMsgQueue_recycle_set(self._q, True)

	mmresult = _w.midiInAddBuffer(self._midiIn, self._hdr)
	if mmresult: raise _mmerror(mmresult, "midiInAddBuffer", True)
	self._readq = []

    def raw_read(self, timeout=NoTimeout):
        """raw_read([timeout]) -> list
	   All queued input received while the device was running
	   is returned as a list of 3-tuples (triples). Each triple
	   is of the form:
	   	(when, msg, bytes)
	   The 'when' member is a timestamp, measured in miliseconds
	   since the last time start() was called. The 'msg' member is
	   one of:
	   	Data, LongData, MoreData, Error, LongError
	   The 'bytes' member is a sequence of bytes (0x00-0x7f) and
	   contain raw MIDI data.
	   The timeout is specified in miliseconds.
	"""
	if not self.state is Started: self.start()
	assert self.isopen
	assert self.state is Started

	result = _w.MidiMsgQueue_read(self._q, timeout)
	if result is None: raise Timeout()
	return result

    def read(self, timeout=NoTimeout):
	"""read([timeout]) -> (timestamp,[byte,...])
	   Returns the next timestamped MIDI data read from the device."""
	if not self._readq:
	    starttime = time.time() * 1000
	    while not self._readq:  # because some packets may be useless
		if timeout == NoTimeout:
		    timeleft = NoTimeout
		else:
		    timeleft = int(starttime + timeout - (time.time() * 1000))
		    if timeleft < 0: timeleft = 0
		packets = self.raw_read(timeleft)
		for msg,when,bytes in packets:
		    if msg in (Data, LongData, MoreData):
			self._readq.insert(0, (when,bytes))
		    elif msg in (Error, LongError):
			if self.nocheck:
			    self._readq.insert(0, (when,bytes))
		    #-- all non-data packets are dropped
	return self._readq.pop()

    def close(self):
	"""close() -> None
	   Resets and closes this device. All resources are released."""
	if not self.isopen: return

	if self._midiIn and self._q:
	    _w.MidiMsgQueue_recycle_set(self._q, False)
	    _w.midiInReset(self._midiIn) # result ignored
	    self.state = Stopped

	if self._q:
	    self._q = None

	self.isopen = False
	self.state = Closed
	try:
	    hdr, self._hdr = self._hdr, None
	    mmresult = _w.midiInUnprepareHeader(self._midiIn, hdr) 
	    if mmresult: raise _mmerror(mmresult, "midiInUnprepareHeader", True)
	finally:
	    midiIn, self._midiIn = self._midiIn, None
	    mmresult = _w.midiInClose(midiIn)
	    if mmresult: raise _mmerror(mmresult, "midiInClose", True)

    def get_event_handle(self):
	"""get_event_handle() -> int
	   Returns the internal Win32 HANDLE that is signaled when
	   midi input is ready. This is meant for use with raw_read().
	   For read(), you should always call read(0) first to see if there
	   is data ready (it will Timeout only if there is none) and only
	   when there isn't then you should wait on this handle."""
	assert self.isopen
	return _w.MidiMsgQueue_get_event_handle(self._q)

    def __del__(self):
	"""destructor
	   Ensures that midi handles are closed properly."""
	if self._q:
	    self._q = None
	if self._hdr and self._midiIn:
	    _w.MidiMsgQueue_recycle_set(self._q, False)
	    _w.midiInReset(self._midiIn) # result ignored
	    _w.midiInUnprepareHeader(self._midiIn, self._hdr) # result ignored
	    self._hdr = None
	if self._midiIn:
	    _w.midiInClose(self._midiIn) # result ignored
	    self.midiIn = None

class OutputDevice:
    """A MIDI output device.
       An output device has the following properties:
	   .id		    - the system index
           .name	    - the product name 
	   .manufacturer_id - the manufacturer's ID
	   .product_id      - the manufacturer's product ID
	   .driver_version  - the (major,minor) tuple of the installed driver
	   .supported 	    - a tuple of zero or more supported capabilities:
	   			Cache, LRVolume, Stream, Volume
	   .voices	    - the number of voices supported (or 0 if n/a)
	   .channels	    - a tuple of channel numbers (1-16) that are 
	   		      available
	   .technology	    - a technology identifier, or an integer
	   			MidiPort, Synth, SQSnyth, FMSynth,
				Mapper, WaveTable, SWSynht
	   .isopen	    - readonly boolean indicating open() was called
	   .state	    - readonly value indicate device soft state
	   			Closed, Stopped, Started
       Calling open() on a device is optional; most methods will call it
       automatically if the isopen flag is false.
    """

    def __init__(self, id, caps):
	#-- use capabilities dictionary to initialise some properties
	self.id =              id
	self.name =            caps['pname']
	self.manufacturer_id = caps['mid']
	self.product_id =      caps['pid']
	self.driver_version =  caps['driverVersion']
	self.voices =          caps['voices']
	_chans = []
	for i in range(16):
	    if caps['channelMask'][i]:
		_chans.append(i+1)
	self.channels = tuple(_chans)
	self.technology =      caps['technology']
	_sup = []
        for nm in (Cache, LRVolume, Stream, Volume):
	    if caps["support"] & getattr(_w, 'MIDICAPS_'+nm.upper()):
		_sup.append(nm)
	self.supported = tuple(_sup)

	self.isopen = False
	self.state = Closed
	self._midiOut = None
	self._hdr = None

    def __repr__(self):
	return "<OutputDevice id %d state %s>" % (
		self.id, self.state)

    def __cmp__(self, other):
	"""Compare two midi ouput devices"""
	if isinstance(other, OutputDevice):
	    return cmp((self.id, self.state), (other.id, other.state))
	else:
	    return cmp(id(self), id(other))

    def open(self, bufsize = 8192):
	"""open([bufsize]) -> None
	   Opens the device ready for writing"""
	assert not self.isopen
	mmresult = _w.midiOutOpen(self.id)
	if type(mmresult) is type(0):
	    raise _mmerror(mmresult, "midiOutOpen", False)
	self._midiOut = mmresult[0]
	self.isopen = True
	self.state = Stopped
	try:
	    self._hdr = _w.new_MIDIHDR(bufsize)
	    mmresult = _w.midiOutPrepareHeader(self._midiOut, self._hdr)
	    if mmresult: raise _mmerror(mmresult, "midiOutPrepareHeader", False)
	except:
	    # close on any error
	    self.close()
	    raise

    def write(self, data):
	"""write(data) -> None
	   Writes the sequence of data bytes out to the midi device."""
	if not self.isopen: self.open()
	assert self.isopen
	if len(data) == 0: 
	    return
	elif len(data) == 1:
	    mmresult = _w.midiOutShortMsg(self._midiOut, 
			    int(data[0]), 0, 0)
	elif len(data) == 2:
	    mmresult = _w.midiOutShortMsg(self._midiOut, 
			    int(data[0]), int(data[1]), 0)
	elif len(data) == 3:
	    mmresult = _w.midiOutShortMsg(self._midiOut, 
			    int(data[0]), int(data[1]), int(data[2]))
	else:
	    _w.MIDIHDR_reset(self._hdr)
	    if _w.MIDIHDR_append(self._hdr, data) == -1:
		raise ValueError, "data too long, or not a sequence of integers"
	    mmresult = _w.midiOutLongMsg(self._midiOut, self._hdr)
	    # XXX Shouldn't this wait until the data is written?
	    # It depends on the driver if it does sync. or async. I/O.
	    # Ideally, a 2nd callback function should be used to signal on 
	    # a MOM_DONE msg and we'd use that to unblock the midiOutLongMsg.

	if mmresult: raise _mmerror(mmresult, "midiOutLongMsg")

    def reset(self):
	"""reset() -> None
	   Resets this device."""
	assert self.isopen
	mmresult = _w.midiOutReset(self._midiOut)
	if mmresult: raise _mmerror(mmresult, "midiOutReset", False)

    def close(self):
	"""close() -> None
	   Closes this device."""
	if not self.isopen: return

	self.isopen = False
	self.state = Closed

	_w.midiOutReset(self._midiOut) # result ignored
	if self._hdr:
	    hdr, self._hdr = self._hdr, None
	    mmresult = _w.midiOutUnprepareHeader(self._midiOut, hdr) 
	    if mmresult: 
		raise _mmerror(mmresult, "midiOutUnprepareHeader", True)
	if self._midiOut:
	    midiOut, self._midiOut = self._midiOut, None
	    mmresult = _w.midiOutClose(midiOut)
	    if mmresult: raise _mmerror(mmresult, "midiOutClose", True)

    def __del__(self):
	"""destructor
	   Ensures that midi handles are closed properly."""
	if self._hdr and self._midiOut:
	    _w.midiOutReset(self._midiOut) # result ignored
	    _w.midiOutUnprepareHeader(self._midiOut, self._hdr) # result ignored
	    self._hdr = None
	if self._midiOut:
	    _w.midiOutClose(self._midiOut) # result ignored
	    self.midiOut = None

def get_input_device(id):
    """get_input_device(id) -> InputDevice
       Constructs a new object attached to the system MIDI input device 
       indicated by the 'id' number"""
    mmresult, caps = _w.midiInGetDevCaps(id)
    if mmresult: raise _mmerror(mmresult, "midiInGetDevCaps", True)
    return InputDevice(id, caps)

def count_input_devices():
    """count_input_devies() -> int
       Returns the number of MIDI input devices in the system.
       The devices are numbered consecutively starting from zero."""
    return _w.midiInGetNumDevs()

def get_input_devices():
    """get_input_devices() -> list
       Constructs a list of all InputDevice objects available"""
    return map(get_input_device, range(count_input_devices()))

def get_output_device(id):
    """get_output_device(id) -> OutputDevice
       Constructs a new object attached to the system MIDI output device 
       indicated by the 'id' number"""
    mmresult, caps = _w.midiOutGetDevCaps(id)
    if mmresult: raise _mmerror(mmresult, "midiOutGetDevCaps", False)
    return OutputDevice(id, caps)

def count_output_devices():
    """count_output_devies() -> int
       Returns the number of MIDI output devices in the system.
       The devices are numbered consecutively starting from zero."""
    return _w.midiOutGetNumDevs()

def get_output_devices():
    """get_output_devices() -> list
       Constructs a list of all OutputDevice objects available"""
    return map(get_output_device, range(count_output_devices()))

def find_input_device(manufacturer_id=None, product_id=None, supported=()):
    """find_input_device([manufacturer_id=int],[product_id=int],
                         [supported=tuple]) -> InputDevice
       Returns the first matching input device, or raises a MatchError"""
    for dev in get_input_devices():
	if manufacturer_id is not None and \
	   dev.manufacturer_id != manufacturer_id:
		continue
	if product_id is not None and \
	   dev.product_id != product_id:
		continue
	supok = False
	for sup in supported:
	    if sup not in dev.supported:
	        break
	else:
	    supok = True
	if not supok: continue
	return dev
    raise NoMatch()

def find_output_device(manufacturer_id=None, product_id=None, supported=(),
		       voices=None, channels=(), technology=None):
    """find_output_device([manufacturer_id=int],[product_id=int],
                          [supported=tuple],[voices=int],[channels=(int,...)],
			  [technology=string]) -> OutputDevice
       Returns the first matching output device, or raises a MatchError"""
    for dev in get_output_devices():
	if manufacturer_id is not None and \
	   dev.manufacturer_id != manufacturer_id:
		continue
	if product_id is not None and \
	   dev.product_id != product_id:
		continue
	supok = False
	for sup in supported:
	    if sup not in dev.supported:
	        break
	else:
	    supok = True
	if not supok: continue
	if voices is not None and \
	   dev.voices < voices:
		continue
	chanok = False
	for chan in channels:
	    if chan not in dev.channels:
	        break
	else:
	    chanok = True
	if technology is not None and \
	   dev.technology != technology:
		continue
	return dev
    raise NoMatch()

