/*
 * Helper functions, included from winmm.i
 */

/* Returns (result, {dictionary}) */
static PyObject *
XmidiInGetDevCaps(UINT id)
{
    MIDIINCAPS caps;
    MMRESULT result;

    result = midiInGetDevCaps(id, &caps, sizeof caps);
    if (result != MMSYSERR_NOERROR)
	return Py_BuildValue("(i,{})", result);
    return Py_BuildValue("(i,{si,si,s(ii),ss,si})",
	    result, 
	    "mid", caps.wMid,
	    "pid", caps.wPid,
	    "driverVersion", 
	        HIBYTE(caps.vDriverVersion),
	        LOBYTE(caps.vDriverVersion),
	    "pname", caps.szPname,
	    "support", caps.dwSupport);
}

static PyObject *
XmidiOutGetDevCaps(UINT id)
{
    MIDIOUTCAPS caps;
    MMRESULT result;

    result = midiOutGetDevCaps(id, &caps, sizeof caps);
    if (result != MMSYSERR_NOERROR)
	return Py_BuildValue("(i,{})", result);
    return Py_BuildValue("(i,{si,si,s(ii),ss,si,si,si,si,s[iiiiiiiiiiiiiiii]})",
	    result, 
	    "mid", caps.wMid,
	    "pid", caps.wPid,
	    "driverVersion", 
	        HIBYTE(caps.vDriverVersion),
	        LOBYTE(caps.vDriverVersion),
	    "pname", caps.szPname,
	    "support", caps.dwSupport,
	    "technology", caps.wTechnology,
	    "voices", caps.wVoices,
	    "notes", caps.wNotes,
	    "channelMask",
	        caps.wChannelMask & (1<<0) ? 1 : 0,
	        caps.wChannelMask & (1<<1) ? 1 : 0,
	        caps.wChannelMask & (1<<2) ? 1 : 0,
	        caps.wChannelMask & (1<<3) ? 1 : 0,
	        caps.wChannelMask & (1<<4) ? 1 : 0,
	        caps.wChannelMask & (1<<5) ? 1 : 0,
	        caps.wChannelMask & (1<<6) ? 1 : 0,
	        caps.wChannelMask & (1<<7) ? 1 : 0,
	        caps.wChannelMask & (1<<8) ? 1 : 0,
	        caps.wChannelMask & (1<<9) ? 1 : 0,
	        caps.wChannelMask & (1<<10) ? 1 : 0,
	        caps.wChannelMask & (1<<11) ? 1 : 0,
	        caps.wChannelMask & (1<<12) ? 1 : 0,
	        caps.wChannelMask & (1<<13) ? 1 : 0,
	        caps.wChannelMask & (1<<14) ? 1 : 0,
	        caps.wChannelMask & (1<<15) ? 1 : 0
	    );
}

static MIDIHDR *
new_MIDIHDR(DWORD length) {
    MIDIHDR *hdr = (MIDIHDR *)malloc(sizeof *hdr);
    /* Round up the length to an integral number of DWORDs */
    if (length % sizeof (DWORD))
	length += sizeof (DWORD) - (length % sizeof (DWORD));
    hdr->lpData = malloc(length);
    hdr->dwBufferLength = length;
    hdr->dwUser = (DWORD)Py_None; Py_INCREF(Py_None);
    hdr->dwFlags = 0;
    return hdr;
}

static void
delete_MIDIHDR(MIDIHDR *self) {
    Py_DECREF((PyObject *)self->dwUser);
    free(self->lpData);
    free(self);
}

/* Empty the buffer */
static void
MIDIHDR_reset(MIDIHDR *self) {
    self->dwBytesRecorded = 0;
    self->dwOffset = 0;
}

/*
 * Append bytes into the buffer. Returns number of bytes appended, 
 * or -1 on error
 */
static int
MIDIHDR_append(MIDIHDR *self, PyObject *src) {
    int srclen, i;

    if (!PySequence_Check(src)) 
	return -1;
    srclen = PySequence_Length(src);
    if (self->dwBytesRecorded + srclen > self->dwBufferLength)
	return -1;
    /* Assign integers */
    for (i = 0; i < srclen; i++) {
	PyObject *item, *intval;
        item = PySequence_GetItem(src, i);
	if (!item) return -1;
	intval = PyNumber_Int(item); Py_DECREF(item);
	if (!intval) return -1;
	self->lpData[self->dwBytesRecorded + i] = 0xff & PyInt_AsLong(intval);
	Py_DECREF(intval);
    }
    self->dwBytesRecorded += srclen;
    return srclen;
}

typedef struct MidiMsgQueue {
	HANDLE	event;		/* set by cb when q is non-empty */
	CRITICAL_SECTION monitor;
	UINT	overflowed;	/* set by cb when it finds q is full */
	UINT	recycle;	/* cb will re-attach buffer when this is true */
	UINT	rpos, wpos;	/* queue read & write pointers */
	UINT	size;		/* size of the buffer in DWORDs */
	DWORD * queue;		/* buffer */
} MidiMsgQueue;

/* Allocate a new queue object */
static MidiMsgQueue *
new_MidiMsgQueue(UINT size)
{
    MidiMsgQueue *q = malloc(sizeof *q);
    if (q) {
	InitializeCriticalSection(&q->monitor);
	q->event = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (q->event == NULL) {
	    PyErr_SetString(PyExc_OSError, "CreateEvent()");
	    free(q);
	    return NULL;
	}
	q->overflowed = FALSE;
        q->recycle = TRUE;
	q->rpos = 0;
	q->wpos = 0;
	q->size = size;
	q->queue = malloc(size * sizeof (DWORD));
	if (!q->queue) {
	    PyErr_NoMemory();
	    free(q);
	    q = NULL;
	}
    }
    return q;
}

/* Delete a queue object (it better not be in use by an open dev!!!) */
static void
delete_MidiMsgQueue(MidiMsgQueue *q)
{
    CloseHandle(q->event);
    DeleteCriticalSection(&q->monitor);
    free(q->queue);
    free(q);
}

/* Provide access to the event handle for integration with other modules */
static long
MidiMsgQueue_get_event_handle(MidiMsgQueue *q)
{
    return (long)q->event;
}

/* Read all the DWORDs from the queue and return as a list. Can block. */
static PyObject *
MidiMsgQueue_read(MidiMsgQueue *q, DWORD timeout)
{
    PyObject *res;
    UINT len, i, j;
    DWORD waitresult;
    BOOL overflowed;

    /* Wait for the queue to become non-empty */
#if 0
    EnterCriticalSection(&q->monitor);
    len = (q->wpos - q->rpos + q->size) % q->size;
    LeaveCriticalSection(&q->monitor);
    if (len == 0)
#endif
    {
        Py_BEGIN_ALLOW_THREADS
           waitresult = WaitForSingleObject(q->event, timeout);
        Py_END_ALLOW_THREADS
	switch (waitresult) {
	case WAIT_TIMEOUT:
	    /* Timed out: return None */
	    Py_INCREF(Py_None);
	    return Py_None;
	default:
	    /* Unexpected wait result */
	    PyErr_SetString(PyExc_OSError, "WaitForSingleObject()");
	    return NULL;
	case WAIT_OBJECT_0:
	    /* Event was signalled as expected. Fall through... */
	    break;
	}
    }

    /* Determine the amount of data we can read from the queue */
    Py_BEGIN_ALLOW_THREADS
    EnterCriticalSection(&q->monitor);
       len = (q->wpos - q->rpos + q->size) % q->size;
       overflowed = q->overflowed;
       q->overflowed = FALSE;
    LeaveCriticalSection(&q->monitor);
    Py_END_ALLOW_THREADS

    /*
     * Construct a list of tuples to return
     */
    res = PyList_New(0);
    i = 0;
    while (i < len + 3) {
	DWORD msg, when, nbytes, pktwords;
	PyObject *tuple, *data;

	msg = q->queue[(q->rpos + i + 0) % q->size];
	when = q->queue[(q->rpos + i + 1) % q->size];
	nbytes = q->queue[(q->rpos + i + 2) % q->size];

	pktwords = 3 + (nbytes + sizeof (DWORD) - 1) / sizeof (DWORD);

    // printf("*c nbytes=%ld\n", nbytes);
    // printf("*c pktwords=%ld\n", pktwords);

	data = PyTuple_New(nbytes);
	for (j = 0; j < nbytes; j++) {
	    DWORD *dw = &q->queue[(q->rpos + i + 3 + (j / sizeof (DWORD))) % q->size];
	    unsigned char *bp = (unsigned char *)dw + (j % sizeof (DWORD));
	    PyTuple_SET_ITEM(data, j, PyInt_FromLong(*bp));
	}


	tuple = PyTuple_New(3);
	PyTuple_SET_ITEM(tuple, 0, PyInt_FromLong(msg));
	PyTuple_SET_ITEM(tuple, 1, PyInt_FromLong(when));
	PyTuple_SET_ITEM(tuple, 2, data);

	PyList_Append(res, tuple);
	Py_DECREF(tuple);		/* because Append does an INCREF */

	i += pktwords;
    }
    // len = i;				/* skip bad bytes? */

    /* If it overflowed, add a None element to the list */
    if (overflowed)
	PyList_Append(res, Py_None);

    /* Advance the read pointer of the queue */
    Py_BEGIN_ALLOW_THREADS
    EnterCriticalSection(&q->monitor);
       q->rpos = (q->rpos + len) % q->size;
    LeaveCriticalSection(&q->monitor);
    Py_END_ALLOW_THREADS

    return res;
}

/*
 * Handle a Midi-input event. The message is carefully inserted into the
 * MidiMsgQueue. If there is room, the overflowed flag is set. Each
 * entry is encoded into the queue as number of DWORDs:
 *     <message> <when> <number-of-bytes> <data>...
 * The data contains the 8-bit byte data (the last DWORD is padded).
 */
void CALLBACK
XMidiInProc(HANDLE midiIn, UINT msg, DWORD instance, DWORD p1, DWORD p2)
{
    char *data;
    DWORD datalen;	/* in bytes */
    DWORD pktwords;
    DWORD maxwords;
    DWORD when;
    MidiMsgQueue *q = (MidiMsgQueue *)instance;
    MIDIHDR *hdr = NULL;
    BOOL overflowed;
    char msgbuf[3];
    UINT i;

    switch (msg) {
    default:
	return;	/* ignore other messages */
    case MIM_DATA:
    case MIM_ERROR:
    case MIM_MOREDATA:
	data = msgbuf;
	when = p2;
	data[0] = LOBYTE(LOWORD(p1));  /* midi status */
	data[1] = HIBYTE(LOWORD(p1));  /* 1st byte of status data */
	data[2] = LOBYTE(HIWORD(p1));  /* 2nd byte of status data */
	datalen = 3;
	break;
    case MIM_LONGDATA:
    case MIM_LONGERROR:
	when = p2;
	hdr = (MIDIHDR *)p1;
	data = hdr->lpData;
	datalen = hdr->dwBytesRecorded;
	break;
    }

    /*
     * The required space in the queue is 
     *    DWORD msg  (MIM_DATA etc)
     *    DWORD when
     *    DWORD number-of-bytes
     *    BYTE  midi-data[]
     *    BYTE  padding[]
     */
    pktwords = 3 + (datalen + sizeof (DWORD) - 1) / sizeof (DWORD);

    /*
     * Check for an overflow condition.
     */
    EnterCriticalSection(&q->monitor);
       /* compute the amount of space remaining in the queue */
       maxwords = (q->size + q->rpos - q->wpos - 1) % q->size;
       if (maxwords < pktwords)
	   q->overflowed = TRUE;
       overflowed = q->overflowed;
    LeaveCriticalSection(&q->monitor);

    if (!overflowed) {
	/* Copy the message and any data into the queue */
	q->queue[(q->wpos + 0) % q->size] = msg; 	/* message */
	q->queue[(q->wpos + 1) % q->size] = when;	/* when */
	q->queue[(q->wpos + 2) % q->size] = datalen;
	for (i = 0; i < pktwords - 3; i++)
	    q->queue[(q->wpos + 3 + i) % q->size] = ((DWORD *)data)[i];

	/* Advance theMMSYSERR_NOERROR write pointer */
	EnterCriticalSection(&q->monitor);
	    q->wpos = (q->wpos + pktwords) % q->size;
	LeaveCriticalSection(&q->monitor);
    }

    if (hdr != NULL && q->recycle) {
	/* Return the used buffer to the system's buffer queue */
        /* XXX It is really naughty to do this here! */
	hdr->dwBytesRecorded = 0;
	midiInAddBuffer(midiIn, hdr, sizeof *hdr);
    }

    /* Signal anyone waiting for the event */
    SetEvent(q->event);
}

/* Wrap midiInOpen() to use the MidiMsgQueue (which may be NULL/None) */
static HMIDIIN
XmidiInOpen(MMRESULT *result, UINT id, MidiMsgQueue *q)
{
    HMIDIIN midiIn = NULL;
    if (q == NULL)
	*result = midiInOpen(&midiIn, id, 0, 0, CALLBACK_NULL);
    else {
	*result = midiInOpen(&midiIn, id, 
		(DWORD)XMidiInProc, (DWORD)q, 
		CALLBACK_FUNCTION | MIDI_IO_STATUS);
    }
    return midiIn;
}

/* Wrap midiOutOpen() to use no callbacks */
static HMIDIOUT
XmidiOutOpen(MMRESULT *result, UINT id)
{
    HMIDIOUT midiOut = NULL;
    *result = midiOutOpen(&midiOut, id, 0, 0, CALLBACK_NULL);
    return midiOut;
}

/* wrap midiOutShortMsg so that the bytes are packed correctly */
static MMRESULT
XmidiOutShortMsg(HMIDIOUT midiOut, UINT status, UINT parm1, UINT parm2)
{
    DWORD msg;

    msg = MAKELONG(MAKEWORD(status, parm1), MAKEWORD(parm2, 0));
    return midiOutShortMsg(midiOut, msg);
}

#define accessor(structname, fieldname, fieldtype) \
    static fieldtype structname ## _ ## fieldname ## _get ( \
	    structname *self) { return self->fieldname; }
accessor(MIDIHDR, dwBytesRecorded, DWORD)
